@@ -3,7 +3,7 @@
  * triglib.c - trigger handling
  *
  * Copyright © 2007 Canonical Ltd
- * Written by Ian Jackson <ian@chiark.greenend.org.uk>
+ * Written by Ian Jackson <ijackson@chiark.greenend.org.uk>
  * Copyright © 2008-2015 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
@@ -26,7 +26,6 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 
-#include <assert.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -49,11 +48,7 @@ static char *triggersdir, *triggersfilefile;
 static char *
 trig_get_filename(const char *dir, const char *filename)
 {
-	char *path;
-
-	m_asprintf(&path, "%s/%s", dir, filename);
-
-	return path;
+	return str_fmt("%s/%s", dir, filename);
 }
 
 static struct trig_hooks trigh;
@@ -106,17 +101,20 @@ void
 trig_clear_awaiters(struct pkginfo *notpend)
 {
 	struct trigaw *ta;
-	struct pkginfo *aw;
 
-	assert(!notpend->trigpend_head);
+	if (notpend->trigpend_head)
+		internerr("package %s has pending triggers",
+		          pkg_name(notpend, pnaw_always));
 
 	ta = notpend->othertrigaw_head;
 	notpend->othertrigaw_head = NULL;
 	for (; ta; ta = ta->samepend_next) {
+		struct pkginfo *aw;
+
 		aw = ta->aw;
 		if (!aw)
 			continue;
-		LIST_UNLINK_PART(aw->trigaw, ta, sameaw.);
+		LIST_UNLINK_PART(aw->trigaw, ta, sameaw);
 		if (!aw->trigaw.head && aw->status == PKG_STAT_TRIGGERSAWAITED) {
 			if (aw->trigpend_head)
 				pkg_set_status(aw, PKG_STAT_TRIGGERSPENDING);
@@ -189,6 +187,38 @@ invalid:
 	return &tki_unknown;
 }
 
+static const char *
+trig_dump_trigger_options(enum trig_options opts)
+{
+	if (opts == TRIG_NOAWAIT)
+		return "/noawait";
+	return "";
+}
+
+/*
+ * Modifies the trigger string by removing the options at the ‘/’ delimiter
+ * and truncating it with a NUL character.
+ */
+static enum trig_options
+trig_parse_trigger_options(char *trigger)
+{
+	char *slash;
+
+	slash = strchr(trigger, '/');
+	if (slash == NULL)
+		return TRIG_AWAIT;
+
+	*slash++ = '\0';
+	if (strcmp("noawait", slash) == 0)
+		return TRIG_NOAWAIT;
+	if (strcmp("await", slash) == 0)
+		return TRIG_AWAIT;
+
+	/* XXX: Should perhaps warn or error for unknown keywords. */
+
+	return TRIG_AWAIT;
+}
+
 /*
  * Calling sequence is:
  *  trig_activate_start(triggername);
@@ -259,8 +289,7 @@ trk_explicit_start(const char *trig)
 	trk_explicit_activate_done();
 
 	varbuf_reset(&trk_explicit_fn);
-	varbuf_add_str(&trk_explicit_fn, triggersdir);
-	varbuf_add_char(&trk_explicit_fn, '/');
+	varbuf_add_dir(&trk_explicit_fn, triggersdir);
 	varbuf_add_str(&trk_explicit_fn, trig);
 	varbuf_end_str(&trk_explicit_fn);
 
@@ -289,7 +318,6 @@ static void
 trk_explicit_activate_awaiter(struct pkginfo *aw)
 {
 	char buf[1024];
-	struct pkginfo *pend;
 
 	if (!trk_explicit_f)
 		return;
@@ -299,18 +327,11 @@ trk_explicit_activate_awaiter(struct pkginfo *aw)
 		        trk_explicit_fn.buf);
 
 	while (trk_explicit_fgets(buf, sizeof(buf)) >= 0) {
+		struct pkginfo *pend;
 		struct dpkg_error err;
-		char *slash;
-		bool noawait = false;
-		slash = strchr(buf, '/');
-		if (slash && strcmp("/noawait", slash) == 0) {
-			noawait = true;
-			*slash = '\0';
-		}
-		if (slash && strcmp("/await", slash) == 0) {
-			noawait = false;
-			*slash = '\0';
-		}
+		enum trig_options opts;
+
+		opts = trig_parse_trigger_options(buf);
 
 		pend = pkg_spec_parse_pkg(buf, &err);
 		if (pend == NULL)
@@ -318,7 +339,7 @@ trk_explicit_activate_awaiter(struct pkginfo *aw)
 			         "illegal package name '%.250s': %.250s"),
 			       trk_explicit_fn.buf, buf, err.str);
 
-		trig_record_activation(pend, noawait ? NULL : aw,
+		trig_record_activation(pend, opts == TRIG_NOAWAIT ? NULL : aw,
 		                       trk_explicit_trig);
 	}
 }
@@ -337,19 +358,31 @@ trk_explicit_interest_change(const char *trig,  struct pkginfo *pkg,
 	atomic_file_open(file);
 
 	while (trk_explicit_f && trk_explicit_fgets(buf, sizeof(buf)) >= 0) {
-		const char *pkgname = pkgbin_name(pkg, pkgbin, pnaw_nonambig);
-		size_t len = strlen(pkgname);
+		enum trig_options trig_opts;
+		struct pkginfo *pkg_parsed;
+		struct dpkg_error err;
 
-		if (strncmp(buf, pkgname, len) == 0 && len < sizeof(buf) &&
-		    (buf[len] == '\0' || buf[len] == '/'))
+		trig_opts = trig_parse_trigger_options(buf);
+
+		pkg_parsed = pkg_spec_parse_pkg(buf, &err);
+		if (pkg_parsed == NULL)
+			ohshit(_("trigger interest file '%.250s' syntax error; "
+			         "illegal package name '%.250s': %.250s"),
+			       trk_explicit_fn.buf, buf, err.str);
+
+		if (pkg == pkg_parsed &&
+		    (pkgbin == &pkg_parsed->installed ||
+		     pkgbin == &pkg_parsed->available))
 			continue;
-		fprintf(file->fp, "%s\n", buf);
+
+		fprintf(file->fp, "%s%s\n", buf,
+		        trig_dump_trigger_options(trig_opts));
 		empty = false;
 	}
 	if (signum > 0) {
 		fprintf(file->fp, "%s%s\n",
-		        pkgbin_name(pkg, pkgbin, pnaw_nonambig),
-		        (opts == TRIG_NOAWAIT) ? "/noawait" : "");
+		        pkgbin_name(pkg, pkgbin, pnaw_same),
+		        trig_dump_trigger_options(opts));
 		empty = false;
 	}
 
@@ -378,6 +411,8 @@ static const struct trigkindinfo tki_explicit = {
 /*---------- File triggers. ----------*/
 
 static struct {
+	/* cppcheck-suppress[unusedStructMember]:
+	 * False positive, macros from dlist.h use the tail member. */
 	struct trigfileint *head, *tail;
 } filetriggers;
 
@@ -399,12 +434,15 @@ trk_file_interest_change(const char *trig, struct pkginfo *pkg,
                          struct pkgbin *pkgbin, int signum,
                          enum trig_options opts)
 {
-	struct filenamenode *fnn;
+	struct fsys_namenode *fnn;
 	struct trigfileint **search, *tfi;
 
 	fnn = trigh.namenode_find(trig, signum <= 0);
 	if (!fnn) {
-		assert(signum < 0);
+		if (signum >= 0)
+			internerr("lost filename node '%s' for package %s "
+			          "triggered to add", trig,
+			          pkgbin_name(pkg, pkgbin, pnaw_always));
 		return;
 	}
 
@@ -426,7 +464,7 @@ trk_file_interest_change(const char *trig, struct pkginfo *pkg,
 	tfi->samefile_next = *trigh.namenode_interested(fnn);
 	*trigh.namenode_interested(fnn) = tfi;
 
-	LIST_LINK_TAIL_PART(filetriggers, tfi, inoverall.);
+	LIST_LINK_TAIL_PART(filetriggers, tfi, inoverall);
 	goto edited;
 
 found:
@@ -440,7 +478,7 @@ found:
 
 	/* Remove it: */
 	*search = tfi->samefile_next;
-	LIST_UNLINK_PART(filetriggers, tfi, inoverall.);
+	LIST_UNLINK_PART(filetriggers, tfi, inoverall);
 edited:
 	filetriggers_edited = 1;
 }
@@ -463,8 +501,8 @@ trig_file_interests_update(void)
 
 	for (tfi = filetriggers.head; tfi; tfi = tfi->inoverall.next)
 		fprintf(file->fp, "%s %s%s\n", trigh.namenode_name(tfi->fnn),
-		        pkgbin_name(tfi->pkg, tfi->pkgbin, pnaw_nonambig),
-		        (tfi->options == TRIG_NOAWAIT) ? "/noawait" : "");
+		        pkgbin_name(tfi->pkg, tfi->pkgbin, pnaw_same),
+		        trig_dump_trigger_options(tfi->options));
 
 	atomic_file_sync(file);
 	atomic_file_close(file);
@@ -507,26 +545,18 @@ trig_file_interests_ensure(void)
 		        triggersfilefile);
 	}
 
-	push_cleanup(cu_closestream, ~0, NULL, 0, 1, f);
+	push_cleanup(cu_closestream, ~0, 1, f);
 	while (fgets_checked(linebuf, sizeof(linebuf), f, triggersfilefile) >= 0) {
 		struct dpkg_error err;
-		char *slash;
-		enum trig_options trig_opts = TRIG_AWAIT;
+		enum trig_options trig_opts;
+
 		space = strchr(linebuf, ' ');
 		if (!space || linebuf[0] != '/')
 			ohshit(_("syntax error in file triggers file '%.250s'"),
 			       triggersfilefile);
 		*space++ = '\0';
 
-		slash = strchr(space, '/');
-		if (slash && strcmp("/noawait", slash) == 0) {
-			trig_opts = TRIG_NOAWAIT;
-			*slash = '\0';
-		}
-		if (slash && strcmp("/await", slash) == 0) {
-			trig_opts = TRIG_AWAIT;
-			*slash = '\0';
-		}
+		trig_opts = trig_parse_trigger_options(space);
 
 		pkg = pkg_spec_parse_pkg(space, &err);
 		if (pkg == NULL)
@@ -545,14 +575,14 @@ ok:
 void
 trig_file_activate_byname(const char *trig, struct pkginfo *aw)
 {
-	struct filenamenode *fnn = trigh.namenode_find(trig, 1);
+	struct fsys_namenode *fnn = trigh.namenode_find(trig, 1);
 
 	if (fnn)
 		trig_file_activate(fnn, aw);
 }
 
 void
-trig_file_activate(struct filenamenode *trig, struct pkginfo *aw)
+trig_file_activate(struct fsys_namenode *trig, struct pkginfo *aw)
 {
 	struct trigfileint *tfi;
 
@@ -579,7 +609,7 @@ trig_file_activate_parents(const char *trig, struct pkginfo *aw)
 }
 
 void
-trig_path_activate(struct filenamenode *trig, struct pkginfo *aw)
+trig_path_activate(struct fsys_namenode *trig, struct pkginfo *aw)
 {
 	trig_file_activate(trig, aw);
 	trig_file_activate_parents(trigh.namenode_name(trig), aw);
@@ -588,7 +618,7 @@ trig_path_activate(struct filenamenode *trig, struct pkginfo *aw)
 static void
 trig_path_activate_byname(const char *trig, struct pkginfo *aw)
 {
-	struct filenamenode *fnn = trigh.namenode_find(trig, 1);
+	struct fsys_namenode *fnn = trigh.namenode_find(trig, 1);
 
 	if (fnn)
 		trig_file_activate(fnn, aw);
@@ -631,7 +661,10 @@ trig_cicb_interest_change(const char *trig, struct pkginfo *pkg,
 {
 	const struct trigkindinfo *tki = trig_classify_byname(trig);
 
-	assert(filetriggers_edited >= 0);
+	if (filetriggers_edited < 0)
+		internerr("trigger control file for package %s not read",
+		          pkgbin_name(pkg, pkgbin, pnaw_always));
+
 	tki->interest_change(trig, pkg, pkgbin, signum, opts);
 }
 
@@ -691,7 +724,7 @@ trig_parse_ci(const char *file, trig_parse_cicb *interest,
 			return; /* No file is just like an empty one. */
 		ohshite(_("unable to open triggers ci file '%.250s'"), file);
 	}
-	push_cleanup(cu_closestream, ~0, NULL, 0, 1, f);
+	push_cleanup(cu_closestream, ~0, 1, f);
 
 	while ((l = fgets_checked(linebuf, sizeof(linebuf), f, file)) >= 0) {
 		for (cmd = linebuf; c_iswhite(*cmd); cmd++) ;
@@ -787,9 +820,6 @@ trig_incorporate(enum modstatdb_rw cstatus)
 			if (errno != EEXIST)
 				ohshite(_("unable to create triggers state"
 				          " directory '%.250s'"), triggersdir);
-		} else if (chown(triggersdir, 0, 0)) {
-			ohshite(_("unable to set ownership of triggers state"
-			          " directory '%.250s'"), triggersdir);
 		}
 		ur = trigdef_update_start(tduf);
 	}
@@ -818,42 +848,12 @@ trig_incorporate(enum modstatdb_rw cstatus)
 
 /*---------- Default hooks. ----------*/
 
-struct filenamenode {
-	struct filenamenode *next;
-	const char *name;
-	struct trigfileint *trig_interested;
-};
-
-static struct filenamenode *trigger_files;
-
-static struct filenamenode *
-th_simple_nn_find(const char *name, bool nonew)
-{
-	struct filenamenode *search;
-
-	for (search = trigger_files; search; search = search->next)
-		if (strcmp(search->name, name) == 0)
-			return search;
-
-	/* Not found. */
-	if (nonew)
-		return NULL;
-
-	search = nfmalloc(sizeof(*search));
-	search->name = nfstrsave(name);
-	search->trig_interested = NULL;
-	search->next = trigger_files;
-	trigger_files = search;
-
-	return search;
-}
-
 TRIGHOOKS_DEFINE_NAMENODE_ACCESSORS
 
 static struct trig_hooks trigh = {
 	.enqueue_deferred = NULL,
 	.transitional_activate = NULL,
-	.namenode_find = th_simple_nn_find,
+	.namenode_find = th_nn_find,
 	.namenode_interested = th_nn_interested,
 	.namenode_name = th_nn_name,
 };
